//===========================================================================
//
//   SSSS    tt          lll  lll       
//  SS  SS   tt           ll   ll        
//  SS     tttttt  eeee   ll   ll   aaaa 
//   SSSS    tt   ee  ee  ll   ll      aa
//      SS   tt   eeeeee  ll   ll   aaaaa  --  "An Atari 2600 VCS Emulator"
//  SS  SS   tt   ee      ll   ll  aa  aa
//   SSSS     ttt  eeeee llll llll  aaaaa
//
// Copyright (c) 1995-1998 by Bradford W. Mott
//
// See the file "license" for information on usage and redistribution of
// this file, and for a DISCLAIMER OF ALL WARRANTIES.
//
// SndPS2.cxx
//============================================================================

#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <string.h>
extern "C"
{
#include <sys/soundcard.h>
}
#include "SndPS2.hxx"
#include "TIASound.h"

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
SoundPS2::SoundPS2(int volume)	: myMute(false)
{
	// Initialize to impossible values so they will be reset 
	// on the first call to the set method
	myAUDC0 = myAUDC1 = myAUDF0 = myAUDF1 = myAUDV0 = myAUDV1 = 255;

	setupSoundDevice();
	openMixer(volume);

}
 
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
SoundPS2::~SoundPS2()
{
	close(dsp_fd);
	free(fragmentBuffer);

	closeMixer();
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
void SoundPS2::set(Sound::Register reg, uInt8 value)
{

  switch(reg) 
  {
    case AUDC0:
      Update_tia_sound(0x15, value);
      break;
    
    case AUDC1:
      Update_tia_sound(0x16, value);
      break;

    case AUDF0:
      Update_tia_sound(0x17, value);
      break;
    
    case AUDF1:
      Update_tia_sound(0x18, value);
      break;

    case AUDV0:
      Update_tia_sound(0x19, value);
      break;

    case AUDV1:
      Update_tia_sound(0x1A, value);
      break;

    default:
      break;
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
void SoundPS2::mute(bool state)
{
	myMute = state;
}

void SoundPS2::handleTime()
{
	// Create the next fragment to play
	Tia_process(fragmentBuffer, fragmentSize);
	
	audio_buf_info info;
	ioctl(dsp_fd,SNDCTL_DSP_GETOSPACE,&info);
	
	if (info.fragments < 1)
	{
		// this means we would block writing
		return;
	}
	
	if (myMute)
	{
		for (int i=0;i<fragmentSize;i++)
			fragmentBuffer[i] = 0x80;
	}
	else
	{
		// poor to lose resolution, but makes it not pop
		for (int i=0;i<fragmentSize;i++)
			fragmentBuffer[i] = fragmentBuffer[i]/2 + 0x80;
	}
	
	// Write fragment to sound device
	write(dsp_fd, fragmentBuffer, fragmentSize);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
void SoundPS2::openMixer(int changeVolume)
{
	int volume;

	if((mixer_fd = open("/dev/mixer", O_RDWR, 0)) == -1)
	{
		printf("stella-sound: Unable to open /dev/mixer device!\n");
		mixer_fd = 0;
		return;
	}
	
	volume = 0;
	if(ioctl(mixer_fd, MIXER_READ(SOUND_MIXER_PCM), &originalVolume) == -1)
	{
		printf("stella-sound: Unable to read mixer settings!\n");    
		close(mixer_fd);
		mixer_fd = 0;
		return;
	}

	volume = changeVolume | (changeVolume << 8);
	if(ioctl(mixer_fd, MIXER_WRITE(SOUND_MIXER_PCM), &volume) == -1)
	{
		printf("stella-sound: Unable to set new volume!\n"); 
		close(mixer_fd);
		mixer_fd = 0;
		return;
	}
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
void SoundPS2::closeMixer()
{
  if(mixer_fd)
  {
    if(ioctl(mixer_fd, MIXER_WRITE(SOUND_MIXER_PCM), &originalVolume) == -1)
      printf("stella-sound: Unable to set original volume!\n");

    close(mixer_fd);
  }
}

void SoundPS2::setupSoundDevice()
{
  fragmentBuffer = NULL;

  /* Open the sound device for writing */
  if((dsp_fd = open("/dev/dsp", O_WRONLY, 0)) == -1)
  {
    printf("stella-sound: Unable to open /dev/dsp device!\n");
    return;
  }

  /* Set the AUDIO DATA FORMAT */
  int format = AFMT_U8;
  if(ioctl(dsp_fd, SNDCTL_DSP_SETFMT, &format) == -1)
  {
    printf("stella-sound: Unable to set 8-bit sample mode!\n");
    return;
  }

  if(format != AFMT_U8)
  {
    printf("stella-sound: Sound card doesn't support 8-bit sample mode!\n");
    return;
  }

  /* Set MONO MODE */
  int stereo = 0;
  if(ioctl(dsp_fd, SNDCTL_DSP_STEREO, &stereo) == -1)
  {
    printf("stella-sound: Sound card doesn't support mono mode!\n");
    return;
  }
 
  if(stereo != 0)
  {
    printf("stella-sound: Sound card doesn't support mono mode!\n");
    return;
  }

  /* Set the SAMPLE RATE */
  sampleRate = 31400;
  if(ioctl(dsp_fd, SNDCTL_DSP_SPEED, &sampleRate) == -1)
  {
    printf("stella-sound: Unable to set sample rate for /dev/dsp!\n");
    return;
  }

  /* Set the NUMBER AND SIZE OF FRAGMENTS */
  int numberAndSizeOfFragments = 0x00020000 | (computeFragmentSize(sampleRate));
  if(ioctl(dsp_fd, SNDCTL_DSP_SETFRAGMENT, &numberAndSizeOfFragments) == -1)
  {
    printf("stella-sound: Unable to set fragment size!\n");
    return;
  }

  /* Query for the actual fragment size */
  ioctl(dsp_fd, SNDCTL_DSP_GETBLKSIZE, &fragmentSize);

  /* Allocate fragment buffer */
  fragmentBuffer = (unsigned char*)malloc(fragmentSize);
  
	/* Initialize the TIA Sound Library */
  Tia_sound_init(31400, sampleRate);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
unsigned long SoundPS2::computeFragmentSize(int sampleRate)
{
  int t;

  for(t = 7; t < 24; ++t) 
  {
    if((1 << t) > (sampleRate / 60))
      return t - 1;
  }

  /* Default to 256 byte fragment size */
  return 8;
}

