//============================================================================
//
//   SSSS    tt          lll  lll       
//  SS  SS   tt           ll   ll        
//  SS     tttttt  eeee   ll   ll   aaaa 
//   SSSS    tt   ee  ee  ll   ll      aa
//      SS   tt   eeeeee  ll   ll   aaaaa  --  "An Atari 2600 VCS Emulator"
//  SS  SS   tt   ee      ll   ll  aa  aa
//   SSSS     ttt  eeeee llll llll  aaaaa
//
// Copyright (c) 1995-2002 by Bradford W. Mott
//
// See the file "license" for information on usage and redistribution of
// this file, and for a DISCLAIMER OF ALL WARRANTIES.
//
// mainPS2.cxx
// created by Chris Kirmse, December 2002, based on the linux-sdl version
//============================================================================

#include <fstream>
#include <iostream>
#include <string>
#include <algorithm>
#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <vector>

#include "bspf.hxx"
#include "Console.hxx"
#include "Event.hxx"
#include "MediaSrc.hxx"
#include "PropsSet.hxx"
#include "System.hxx"
#include "SndPS2.hxx"
#include "Settings.hxx"

//ps2 stuff
#include "PS2Pad.h"

#include "texture.h"
#include "primitives.h"
#include "PS2Utils.h"
#include "menu.h"
#include "filemenu.h"
#include "pausemenu.h"
#include "collisioneffect.h"

#include "mainPS2.h"

// function prototypes

static void SetupDisplay(int crt_mode);
static void SetupJoysticks();

static bool LoadROMImage(const char *file);
static void SetupNewGameDisplay();
static void SetupPalette();
static void cleanup();

static void UpdateGameDisplay(MediaSource& mediaSource);
static void CheckEnhancements();
static void SetState(int new_state);
static int AnimateInterpolate(int source,int dest,int current_time,int total_time);
static void RunTime(int micros);
static void DrawEverything();
static void HandleInput();
static void HandleGamePausedInput();
static void HandleGameInput();

static void Quit();
static void togglePause();
static uInt32 getTicks();

static bool setupProperties(PropertiesSet& set);
static void handleRCFile();

static void ParseCommandLine(int argc,char *argv[]);

// Event objects to use
static Event theEvent;

static SoundPS2 *sound = NULL;

// Pointer to the console object or the null pointer
static Console* theConsole = (Console*) NULL;

// Pointer to the settings object or the null pointer
static Settings* settings = (Settings*) NULL;

// Indicates if the user wants to quit
static bool quit = false;

// Indicates if the emulator should be paused
static bool is_paused = false;

// whether or not we draw the game screen bilinearly filtered
static bool filtered = false;

static int crt_mode;

static Texture game_texture;
static QuadTex game_quad;

static Texture background_texture;
static QuadTex background_quad;

static Texture font_texture;
static Font font;

static FileMenu *file_menu = NULL;
static bool current_game_selected = false;

static PauseMenu *pause_menu = NULL;

enum
{
	STATE_NONE = 0,
	STATE_UI = 1,
	STATE_ANIMATING_TO_GAME = 2,
	STATE_IN_GAME = 3,
	STATE_ANIMATING_TO_UI = 4,
};
static int state = STATE_NONE;

// ui data
static const int UI_GAME_X = -20;
static const int UI_GAME_Y = -60;
static const int UI_GAME_W = 320;
static const int UI_GAME_H = 210;

// game data
static const int GAME_X = -320;
static const int GAME_Y = -222;
static const int GAME_W = 640;
static const int GAME_H = 448;

static bool draw_memory_info = false;

// animating_to_game or animate_to_ui data
static const int ANIMATE_TOTAL_MICROS = 200000;
static int animate_progress_micros = 0;

// all the x, y, w, h are based on 640x448; need to scale
// in interlaced video modes
inline int ScaleX(int source_x)
{
	return source_x;
}

inline int ScaleY(int source_y)
{
	if (crt_mode == PS2_GS_NTSC || crt_mode == PS2_GS_PAL)
		return source_y/2;
	return source_y;
}	

bool GetFiltered()
{
	return filtered;
}

void SetFiltered(bool new_filtered)
{
	filtered = new_filtered;
	SetQuadTexFilter(&game_quad,filtered);
}

Event * GetEvent()
{
	return &theEvent;
}

void GoToUI()
{
	SetState(STATE_ANIMATING_TO_UI);
}

void SetupDisplay(int crt_mode)
{
	DisplayInit(crt_mode);
	DisplayFlip();
	DisplayFlip();

	game_texture.Create8BitPalettized(256,256);
	for (int i=0;i<256;i++)
	{
		memset(game_texture.image_data + i*256,i,256);
		game_texture.SetPaletteEntry(i,i,i,i,128);
	}
	game_texture.ResendPaletteToGS();
	game_texture.ResendImageToGS();

	InitQuadTex(&game_quad,0,0,0,0,0,filtered,&game_texture);

	background_texture.LoadBitmap("background.bmp");
	InitQuadTex(&background_quad,ScaleX(GAME_X),ScaleY(GAME_Y),
					ScaleX(GAME_W),ScaleY(GAME_H),0,1,&background_texture);
	SetQuadTexUV(&background_quad,0,0,512,384);

	font_texture.LoadBitmap("courier8.bmp");
	InitFont(&font,&font_texture,-6);

	// setup UI
	file_menu = new FileMenu();
	file_menu->SetPosition(ScaleX(-290),ScaleY(-110));
	file_menu->SetSize(ScaleX(250),ScaleY(290));
	file_menu->SetFont(&font);

	pause_menu = new PauseMenu();
	pause_menu->SetPosition(ScaleX(-160),ScaleY(-60));
	pause_menu->SetSize(ScaleX(320),ScaleY(120));
	pause_menu->SetFont(&font);

	SetState(STATE_UI);
}

void SetupSound()
{
	// Create a sound object for use with the console
	sound = new SoundPS2(settings->theDesiredVolume);
}

void SetupJoysticks()
{
	PS2PadManager::Initialize();
}

bool LoadROMImage(const char *file)
{
	if (theConsole != NULL)
	{
		delete theConsole;
		theConsole = NULL;
	}

	// Open the cartridge image and read it in
	ifstream in(file); 
	if(!in)
	{
		cerr << "ERROR: Couldn't open " << file << "..." << endl;
		return false;
	}

	uInt8* image = new uInt8[512 * 1024];
	in.read((char*)image, 512 * 1024);
	uInt32 size = in.gcount();
	in.close();

	// Create a properties set for us to use and set it up
	PropertiesSet propertiesSet;
	if(!setupProperties(propertiesSet))
	{
		delete[] image;
		return false;
	}
	
	// Get just the filename of the file containing the ROM image
	const char* filename = (!strrchr(file, '/')) ? file : strrchr(file, '/') + 1;

	// Create the 2600 game console
	theConsole = new Console(image, size, filename, 
									 theEvent, propertiesSet, *sound);

	// Free the image since we don't need it any longer
	delete[] image;

	// always unpaused on loading
	if (is_paused)
	{
		togglePause();
		sound->mute(false);
	}

	return true;
}

static void SetupNewGameDisplay()
{
	SetQuadTexUV(&game_quad,0,0,theConsole->mediaSource().width(),
					 theConsole->mediaSource().height());

	SetupPalette();
}

/**
  Recalculates palette of an 8-bit (256 color) screen.
*/
void SetupPalette()
{
	// Make the palette be half-bright if pause is selected
	uInt8 shift = 0;
	if(is_paused)
		shift = 1;

	Byte *clut_data = game_texture.clut_data;

	// Map 2600 colors to the current screen
	const uInt32* gamePalette = theConsole->mediaSource().palette();
	for(uInt32 i = 0; i < 256; ++i)
	{
		uInt8 r, g, b;
		r = (uInt8) ((gamePalette[i] & 0x00ff0000) >> 16) >> shift;
		g = (uInt8) ((gamePalette[i] & 0x0000ff00) >> 8) >> shift;
		b = (uInt8) (gamePalette[i] & 0x000000ff) >> shift;

		game_texture.SetPaletteEntry(i,r,g,b,128);
	}

	game_texture.ResendPaletteToGS();
}

/**
  This routine is called when the program is about to quit.
*/
void Quit()
{
	quit = true;
}

/**
  Toggles pausing of the emulator
*/
void togglePause()
{
	is_paused = !is_paused;

  // Pause the console
  theConsole->mediaSource().pause(is_paused);

  // Show a different palette depending on pause state
  SetupPalette();
}

// this handles updating for a new emulated game bitmap
void UpdateGameDisplay(MediaSource& mediaSource)
{
	uInt8* currentFrame = mediaSource.currentFrameBuffer();
	uInt8* previousFrame = mediaSource.previousFrameBuffer();

	int height = mediaSource.height();
	int width = mediaSource.width();

	// inefficient--rendered into main memory, copied to main memory here,
	// then copied to texture memory
	for (int i=0;i<height;i++)
		memcpy(game_texture.image_data + i*256,currentFrame + i*width,width);

	game_texture.ResendImageToGS();
}

void CheckEnhancements()
{
	if (draw_memory_info)
	{
		char buffer[200];
	
		// dump tia registers
		buffer[0] = 0;
		for (int j=0;j<16;j++)
		{
			uInt8 x = theConsole->system().peek(j);
			sprintf(&buffer[strlen(buffer)],"%02x ",(unsigned int)x);
		}
		DrawString(buffer,ScaleX(-320),ScaleY(-200),&font,true);
		
		// dump m6532 registers
		buffer[0] = 0;
		for (int j=0;j<8;j++)
		{
			uInt8 x = theConsole->system().peek(0x280 + j);
			sprintf(&buffer[strlen(buffer)],"%02x ",(unsigned int)x);
		}
		DrawString(buffer,ScaleX(-320),ScaleY(-180),&font,true);

		// dump ram
		for (int i=0;i<8;i++)
		{
			buffer[0] = 0;
			for (int j=0;j<16;j++)
			{
				uInt8 x = theConsole->system().peek(0x80 + j + i*16);
				sprintf(&buffer[strlen(buffer)],"%02x ",(unsigned int)x);
			}
			DrawString(buffer,ScaleX(-320),ScaleY(-160 + 16*i),&font,true);
		}
	}

	// special addition of force feedback for Combat cartridge
	if (strncmp("Combat",theConsole->properties().get("Cartridge.Name").c_str(),6) != 0)
		return;

	// if not currently playing, then quit
	if (theConsole->system().peek(0x80 + 90) != 0x3c &&
		 theConsole->system().peek(0x80 + 90) != 0x2c &&
		 theConsole->system().peek(0x80 + 90) != 0x9c &&
		 theConsole->system().peek(0x80 + 90) != 0x98)
		return;

	if (theConsole->system().peek(0x80 + 25) == 0x3d)
	{
		// player 1 fired
		PS2Pad *joypad = PS2PadManager::GetPad(0);
		joypad->SetVibrationEffect(new CollisionEffect(9,0,4));
	}

	if (theConsole->system().peek(0x80 + 26) == 0x3d)
	{
		// player 2 fired
		PS2Pad *joypad = PS2PadManager::GetPad(1);
		joypad->SetVibrationEffect(new CollisionEffect(9,0,4));
	}

	if (theConsole->system().peek(0x80 + 42) == 0x80)
	{
		// player 1 moving
		PS2Pad *joypad = PS2PadManager::GetPad(0);
		joypad->SetVibrationEffect(new CollisionEffect(9,160,0));
	}

	if (theConsole->system().peek(0x80 + 43) == 0x80)
	{
		// player 2 moving
		PS2Pad *joypad = PS2PadManager::GetPad(1);
		joypad->SetVibrationEffect(new CollisionEffect(9,160,0));
	}


	static int prev_p1_score = 0;
	int p1_score = theConsole->system().peek(0x80 + 33);
	if (p1_score == prev_p1_score+1)
	{
		// player 2 hit
		PS2Pad *joypad = PS2PadManager::GetPad(1);
		joypad->SetVibrationEffect(new CollisionEffect(60,255,255));
	}
	prev_p1_score = p1_score;

	static int prev_p2_score = 0;
	int p2_score = theConsole->system().peek(0x80 + 34);
	if (p2_score == prev_p2_score+1)
	{
		// player 1 hit
		PS2Pad *joypad = PS2PadManager::GetPad(0);
		joypad->SetVibrationEffect(new CollisionEffect(60,255,255));
	}
	prev_p2_score = p2_score;
}

void SetState(int new_state)
{
	if (state == new_state)
		return;

	state = new_state;

	switch (state)
	{
	case STATE_NONE :
		break;
	case STATE_UI :
		MoveQuadTex(&game_quad,ScaleX(UI_GAME_X),ScaleY(UI_GAME_Y),
						ScaleX(UI_GAME_W),ScaleY(UI_GAME_H));
		break;
	case STATE_ANIMATING_TO_GAME :
		animate_progress_micros = 0;
		break;
	case STATE_IN_GAME :
		MoveQuadTex(&game_quad,ScaleX(GAME_X),ScaleY(GAME_Y),
						ScaleX(GAME_W),ScaleY(GAME_H));
		break;
	case STATE_ANIMATING_TO_UI :
		animate_progress_micros = 0;
		break;
	default :
		printf("invalid state %i\n",state);
	}

}

static int AnimateInterpolate(int source,int dest,int current_time,int total_time)
{
	return source + current_time*(dest-source)/total_time;
}

static void RunTime(int micros)
{
	switch (state)
	{
	case STATE_NONE :
		break;
	case STATE_UI :
		break;
	case STATE_ANIMATING_TO_GAME :
	{
		animate_progress_micros += micros;
		if (animate_progress_micros > ANIMATE_TOTAL_MICROS)
		{
			SetState(STATE_IN_GAME);
			return;
		}
		int x = AnimateInterpolate(ScaleX(UI_GAME_X),ScaleX(GAME_X),
											animate_progress_micros,ANIMATE_TOTAL_MICROS);
		int y = AnimateInterpolate(ScaleY(UI_GAME_Y),ScaleY(GAME_Y),
											animate_progress_micros,ANIMATE_TOTAL_MICROS);
		int w = AnimateInterpolate(ScaleX(UI_GAME_W),ScaleX(GAME_W),
											animate_progress_micros,ANIMATE_TOTAL_MICROS);
		int h = AnimateInterpolate(ScaleY(UI_GAME_H),ScaleY(GAME_H),
											animate_progress_micros,ANIMATE_TOTAL_MICROS);
		MoveQuadTex(&game_quad,x,y,w,h);
		break;
	}
	case STATE_IN_GAME :
		break;
	case STATE_ANIMATING_TO_UI :
	{
		animate_progress_micros += micros;
		if (animate_progress_micros > ANIMATE_TOTAL_MICROS)
		{
			SetState(STATE_UI);
			return;
		}
		int x = AnimateInterpolate(ScaleX(GAME_X),ScaleX(UI_GAME_X),
											animate_progress_micros,ANIMATE_TOTAL_MICROS);
		int y = AnimateInterpolate(ScaleY(GAME_Y),ScaleY(UI_GAME_Y),
											animate_progress_micros,ANIMATE_TOTAL_MICROS);
		int w = AnimateInterpolate(ScaleX(GAME_W),ScaleX(UI_GAME_W),
											animate_progress_micros,ANIMATE_TOTAL_MICROS);
		int h = AnimateInterpolate(ScaleY(GAME_H),ScaleY(UI_GAME_H),
											animate_progress_micros,ANIMATE_TOTAL_MICROS);
		MoveQuadTex(&game_quad,x,y,w,h);
		break;
	}
	default :
		printf("invalid state %i\n",state);
	}

}

void DrawEverything()
{
	switch (state)
	{
	case STATE_NONE :
		break;
	case STATE_UI :
	case STATE_ANIMATING_TO_GAME :
	case STATE_ANIMATING_TO_UI :
		DrawQuadTex(&background_quad);
		DrawQuadTex(&game_quad);
		file_menu->Draw();
		break;

	case STATE_IN_GAME :
		MoveQuadTex(&game_quad,ScaleX(GAME_X),ScaleY(GAME_Y),
						ScaleX(GAME_W),ScaleY(GAME_H));
		DrawQuadTex(&game_quad);
		if (is_paused)
			pause_menu->Draw();
		break;

	default :
		printf("invalid state %i\n",state);
	}
}

/**
  This routine should be called regularly to handle events
*/
void HandleInput()
{
	PS2PadManager::Update();

	PS2Pad *joypad = PS2PadManager::GetPad(0);

	if (joypad->WasButtonPressed(PS2Pad::BUTTON_R3))
		ScreenCapture("stella.bmp");

	if (joypad->WasButtonPressed(PS2Pad::BUTTON_L3))
		draw_memory_info = !draw_memory_info;

	if (state == STATE_IN_GAME)
	{
		if (is_paused)
			HandleGamePausedInput();
		else
			HandleGameInput();
		return;
	}

	if (state == STATE_ANIMATING_TO_GAME || state == STATE_ANIMATING_TO_UI)
		return;

	// main menu
	file_menu->Move(joypad->IsButtonPressed(PS2Pad::BUTTON_UP),
						 joypad->IsButtonPressed(PS2Pad::BUTTON_DOWN));

	if (joypad->IsButtonPressed(PS2Pad::BUTTON_UP) ||
		 joypad->IsButtonPressed(PS2Pad::BUTTON_DOWN))
	{
		current_game_selected = false;
	}

	if (joypad->WasButtonPressed(PS2Pad::BUTTON_CROSS))
	{
		if (current_game_selected)
		{
			SetState(STATE_ANIMATING_TO_GAME);
		}
		else
		{
			bool have_file;
			file_menu->Selected(&have_file);
			if (have_file)
			{
				current_game_selected = true;
				string filename = file_menu->GetSelectedFilename();
				if (LoadROMImage(filename.c_str()) == false)
					return;
				SetupNewGameDisplay();
			}
		}
	}

	if (joypad->IsButtonPressed(PS2Pad::BUTTON_START))
		Quit();
}

void HandleGamePausedInput()
{
	PS2Pad *joypad = PS2PadManager::GetPad(0);

	pause_menu->Move(joypad->IsButtonPressed(PS2Pad::BUTTON_UP),
						  joypad->IsButtonPressed(PS2Pad::BUTTON_DOWN));

	if (joypad->WasButtonPressed(PS2Pad::BUTTON_CROSS))
	{
		pause_menu->Selected();
	}

	if (joypad->WasButtonPressed(PS2Pad::BUTTON_START))
		togglePause();
}

void HandleGameInput()
{
	// controller 1

	PS2Pad *joypad = PS2PadManager::GetPad(0);

	// joystick
	theEvent.set(Event::JoystickZeroLeft,
					 joypad->IsButtonPressed(PS2Pad::BUTTON_LEFT));

	theEvent.set(Event::JoystickZeroRight,
					 joypad->IsButtonPressed(PS2Pad::BUTTON_RIGHT));

	theEvent.set(Event::JoystickZeroUp,
					 joypad->IsButtonPressed(PS2Pad::BUTTON_UP));

	theEvent.set(Event::JoystickZeroDown,
					 joypad->IsButtonPressed(PS2Pad::BUTTON_DOWN));

	theEvent.set(Event::JoystickZeroFire,
					 joypad->IsButtonPressed(PS2Pad::BUTTON_CROSS));

	if (joypad->WasButtonPressed(PS2Pad::BUTTON_START))
		togglePause();

	if (joypad->WasButtonPressed(PS2Pad::BUTTON_TRIANGLE))
		SetState(STATE_ANIMATING_TO_UI);

	theEvent.set(Event::ConsoleSelect,joypad->IsButtonPressed(PS2Pad::BUTTON_SELECT));
	theEvent.set(Event::ConsoleReset,joypad->IsButtonPressed(PS2Pad::BUTTON_R1));

	// paddle
	theEvent.set(Event::PaddleZeroResistance,
					 500000*(1-joypad->GetLeftX()));
	theEvent.set(Event::PaddleZeroFire,
					 joypad->IsButtonPressed(PS2Pad::BUTTON_CROSS));

	// controller 2

	// joystick
	joypad = PS2PadManager::GetPad(1);

	theEvent.set(Event::JoystickOneLeft,
					 joypad->IsButtonPressed(PS2Pad::BUTTON_LEFT));

	theEvent.set(Event::JoystickOneRight,
					 joypad->IsButtonPressed(PS2Pad::BUTTON_RIGHT));

	theEvent.set(Event::JoystickOneUp,
					 joypad->IsButtonPressed(PS2Pad::BUTTON_UP));

	theEvent.set(Event::JoystickOneDown,
					 joypad->IsButtonPressed(PS2Pad::BUTTON_DOWN));

	theEvent.set(Event::JoystickOneFire,
					 joypad->IsButtonPressed(PS2Pad::BUTTON_CROSS));

	// paddle
	theEvent.set(Event::PaddleOneResistance,
					 500000*(1-joypad->GetLeftX()));
	theEvent.set(Event::PaddleOneFire,
					 joypad->IsButtonPressed(PS2Pad::BUTTON_CROSS));
}

/**
  Setup the properties set by first checking for a user file ".stella.pro",
  then a system-wide file "/etc/stella.pro".  Return false if neither file
  is found, else return true.

  @param set The properties set to setup
*/
bool setupProperties(PropertiesSet& set)
{
  string homePropertiesFile = getenv("HOME");
  homePropertiesFile += "/.stella.pro";
  string systemPropertiesFile = "/etc/stella.pro";

  // Check to see if the user has specified an alternate .pro file.
  // If it exists, use it.
  if(settings->theAlternateProFile != "")
  {
    if(access(settings->theAlternateProFile.c_str(), R_OK) == 0)
    {
      set.load(settings->theAlternateProFile, &Console::defaultProperties(), false);
      return true;
    }
    else
    {
      cerr << "ERROR: Couldn't find \"" << settings->theAlternateProFile <<
              "\" properties file." << endl;
      return false;
    }
  }

  if(access(homePropertiesFile.c_str(), R_OK) == 0)
  {
    set.load(homePropertiesFile, &Console::defaultProperties(), false);
    return true;
  }
  else if(access(systemPropertiesFile.c_str(), R_OK) == 0)
  {
    set.load(systemPropertiesFile, &Console::defaultProperties(), false);
    return true;
  }
  else
  {
    set.load("", &Console::defaultProperties(), false);
    return true;
  }
}


/**
  Should be called to determine if an rc file exists.  First checks if there
  is a user specified file ".stellarc" and then if there is a system-wide
  file "/etc/stellarc".
*/
void handleRCFile()
{
	string homeRCFile = getenv("HOME");
	homeRCFile += "/.stellarc";

	if(access(homeRCFile.c_str(), R_OK) == 0 )
	{
		ifstream homeStream(homeRCFile.c_str());
		settings->handleRCFile(homeStream);
	}
	else if(access("/etc/stellarc", R_OK) == 0 )
	{
		ifstream systemStream("/etc/stellarc");
		settings->handleRCFile(systemStream);
	}
}


/**
  Does general cleanup in case any operation failed (or at end of program).
*/
void cleanup()
{
	if (theConsole)
		delete theConsole;

	if (sound)
		delete sound;

	if (settings)
		delete settings;

	if (file_menu)
		delete file_menu;

	if (pause_menu)
		delete pause_menu;

	DisplayDeInit();

	PS2PadManager::Shutdown();
}

void ParseCommandLine(int argc,char *argv[])
{
	crt_mode = PS2_GS_VESA;

	for (int i=1;i<argc;i++)
	{
		if (strcmp(argv[i],"-ntsc") == 0)
			crt_mode = PS2_GS_NTSC;
		if (strcmp(argv[i],"-pal") == 0)
			crt_mode = PS2_GS_PAL;
		if (strcmp(argv[i],"-dtv") == 0)
			crt_mode = PS2_GS_DTV;
		if (strcmp(argv[i],"-vesa") == 0)
			crt_mode = PS2_GS_VESA;
	}
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
int main(int argc, char* argv[])
{
	// First create some settings for the emulator
	settings = new Settings();
	if(!settings)
	{
		cerr << "ERROR: Couldn't create settings." << endl;
		cleanup();
		return 0;
	}

	// Load in any user defined settings from an RC file
	handleRCFile();

	ParseCommandLine(argc,argv);
	
	SetupDisplay(crt_mode);
	SetupSound();
	SetupJoysticks();

	int prev_time = -1;

	// Main game loop
	while (!quit)
	{
		int startTime = getTicks();

		if (theConsole != NULL)
		{
			theConsole->mediaSource().update();
			UpdateGameDisplay(theConsole->mediaSource());
		}

		DrawEverything();

		if (theConsole != NULL)
			CheckEnhancements();

		if (draw_memory_info)
		{
			char buffer[100];
			sprintf(buffer,"micros pf  %i",prev_time);
			DrawString(buffer,ScaleX(-320),ScaleY(-220),&font,true);
		}

		sound->handleTime();
		HandleInput();

		prev_time = getTicks() - startTime;

		DisplayFlip();						

		RunTime(getTicks() - startTime);
	}

	// Cleanup time ...
	cleanup();
	return 0;
}


/**
  Returns number of ticks in microseconds
*/
inline uInt32 getTicks()
{
	timeval now;
	gettimeofday(&now, 0);

	return (uInt32) (now.tv_sec * 1000000 + now.tv_usec);
}
