#include "primitives.h"
#include "fixed_tc.h"

//from main.c
extern int g_fd_gs;		//file descriptor of gs
extern int g_paddata;	//controller button info


//*************************************************************************************
//										PIXEL
//*************************************************************************************

void InitPixel(Pixel *p, int x, int y, int r, int g, int b)
{
	//Here's a small walkthrough
	
	//Set packet.begin pointer to the allocated array memory(regdata)
	InitGifPacket(&(p->packet),(u128 *)p->regdata);
	
	//AddGSData adds 64 bits of data to regdata.  The first parameter is the
	//array index of regdata to write to.  When setting up a primitive, just
	//start with 0 and continue in order.  The point of having the index is
	//so that you can change a value later(like the x,y coordinates) without
	//having to rewrite the whole packet(see MoveQuadTex).
	
	//First, we want to setup the giftag(see p.151 of the EEUSER pdf).  Since
	//the giftag is 128 bits long and AddGSData only adds 64 bits at a time,
	//we'll setup the giftag using 2 calls to AddGSData.  This first call
	//sets up the NLOOP, EOP, PRE, PRIM, FLG, and NREG fields of the giftag.
	//
	//In this case, we want NLOOP = 1  (do everything that follows once)
	//                      EOP   = 1  (this is the end of the packet)
	//                      PRE   = 0  (this says to ignore the PRIM field, 
	//                                  we'll setup the PRIM register later)
	//						PRIM  = 0  (since PRE=0, this is ignored)
	//						FLG   = PS2_GIFTAG_FLG_REGLIST (reglist mode)
	//						NREG  = 3  (we're going to set 3 registers:
	//                                  PRIM, RGBAQ, and XYZ)
	//
	//So that's the first half of the giftag
	AddGSData(0,PS2_GIFTAG_SET_TOPHALF(1,1,0,0,PS2_GIFTAG_FLG_REGLIST,3),&(p->packet));
	
	//Next, we need to setup the REGS field of the giftag.  This is 64 bits
	//so we'll just use AddGSData.  Since we set the FLG field in the first
	//part of the giftag to PS2_GIFTAG_REGLIST, we'll be operating in
	//REGLIST mode(see p.159 of EEUSER pdf).  In this mode, we have to
	//specify which registers we're going to set in the REGS field of the
	//giftag.  To do this, we send a list of the addresses of the registers 
	//into the 64 bits of the REGS field.  We start with the address
	//of the first register we're going to set(PS2_GS_PRIM in this case).
	//Then, we do a bitwise OR with each successive register address, 
	//making sure to shift each one 4 more bits to the left since each address 
	//is 4 bits long.  This builds a string of addresses. 
	//
	//NOTE: The order of the registers is very important.
	//      See p.156-159 of the EEUSER pdf to see examples.  The
	//      examples of PACKED mode basically apply to REGLIST mode, too.
	//
	//In this case, we want to set the PRIM, RGBAQ, and XYZF2 registers since
	//that's all we need to define a pixel.
	//
	//That's all for the giftag.
	AddGSData(1,(PS2_GS_PRIM | PS2_GS_RGBAQ<<4 | PS2_GS_XYZF2<<8),&(p->packet));
	
	//Now that we've setup the giftag, we need to send the actual data that's
	//going to go into the registers we defined.  We need to send
	//it in the order that we setup the REGS field of the giftag to make sure
	//it goes to the right place.  
	//
	//In this case, we need to start with the PRIM register, since that was
	//the first in our REGS list.
	//
	//See p.116 of the GSUSER pdf for more info on the fields of the PRIM
	//register.  Basically, we just tell it we want a POINT primitive since
	//we won't need anything else like texture mapping, alpha blending, etc.
	AddGSData(2,PS2_GS_SETREG_PRIM(PS2_GS_PRIM_PRIM_POINT,0,0,0,0,0,0,0,0),&(p->packet));
	
	//Next in our REGS list was RGBAQ, so send that data.(p.119 of GSUSER)
	//
	//In this case, we just send the r,g,b parameters of this InitPixel
	//function, then 128 for the alpha value(128 is the max), and 0 for
	//the Q value since we won't be using texture mapping.
	AddGSData(3,PS2_GS_SETREG_RGBAQ(r,g,b,128,0),&(p->packet));
	
	//Finally, the XYZ register.(p.139 of GSUSER)
	//
	//First of all, the center of the screen is at coordinates (2048,2048).
	//This can be obtained from gp->center_x(gp is defined in main.c)
	//So we want to add our x,y coordinates to these.  This means that
	//anything on the top half of the screen will have a negative y
	//coordinate and anything on the left half will have a negative x.
	//Also, the coordinates need to have 4 decimal places, so we shift
	//them 4 bits to the left.
	AddGSData(4,PS2_GS_SETREG_XYZ((2048+x)<<4,(2048+y)<<4,0),&(p->packet));
	
	//Finally, terminate the packet.
	TerminateGifPacket(&(p->packet));
	
	//Set a few member variables of the Pixel struct
	p->x = x;
	p->y = y;
	p->packetsize = SizeGifPacket(&(p->packet));
}

void DrawPixel(Pixel *p)
{
	//Send it
	ps2_dma_start_n(g_fd_gs,p->packet.begin,p->packetsize);
}


//*************************************************************************************
//										LINE
//*************************************************************************************

//x0,y0 are the position of the beginning of the line
//x1,y1 are the end

void InitLine(Line *l, int x0, int y0, int x1, int y1, int r, int g, int b)
{
	InitGifPacket(&(l->packet),(u128 *)l->regdata);
	
	AddGSData(0,PS2_GIFTAG_SET_TOPHALF(1,1,0,0,PS2_GIFTAG_FLG_REGLIST,4),&(l->packet));
	AddGSData(1,(PS2_GS_PRIM | PS2_GS_RGBAQ<<4 | PS2_GS_XYZF2<<8 | PS2_GS_XYZF2<<12),&(l->packet));
	AddGSData(2,PS2_GS_SETREG_PRIM(PS2_GS_PRIM_PRIM_LINE,0,0,0,0,0,0,0,0),&(l->packet));
	AddGSData(3,PS2_GS_SETREG_RGBAQ(r,g,b,128,0),&(l->packet));
	AddGSData(4,PS2_GS_SETREG_XYZ((2048+x0)<<4,(2048+y0)<<4,0),&(l->packet));
	AddGSData(5,PS2_GS_SETREG_XYZ((2048+x1)<<4,(2048+y1)<<4,0),&(l->packet));
	TerminateGifPacket(&(l->packet));
	
	l->x0 = x0;
	l->y0 = y0;
	l->x1 = x1;
	l->y1 = y1;
	l->packetsize = SizeGifPacket(&(l->packet));
}

void InitLineStrip(Line *l, int x, int y, int w, int h, int *verts, int numverts, int r, int g, int b)
{
	InitGifPacket(&(l->packet),(u128 *)l->regdata);
	
	AddGSData(0,PS2_GIFTAG_SET_TOPHALF(1,0,0,0,PS2_GIFTAG_FLG_REGLIST,4),&(l->packet));
	AddGSData(1,(PS2_GS_PRIM | PS2_GS_RGBAQ<<4 | PS2_GS_XYZF2<<8 | PS2_GS_XYZF2<<12),&(l->packet));
	AddGSData(2,PS2_GS_SETREG_PRIM(PS2_GS_PRIM_PRIM_LINESTRIP,0,0,0,0,0,0,0,0),&(l->packet));
	AddGSData(3,PS2_GS_SETREG_RGBAQ(r,g,b,128,0),&(l->packet));
	AddGSData(4,PS2_GS_SETREG_XYZ((2048+x)<<4,(2048+y)<<4,0),&(l->packet));
	AddGSData(5,PS2_GS_SETREG_XYZ((2048+x+w)<<4,(2048+y+h)<<4,0),&(l->packet));
	AddGSData(6,PS2_GIFTAG_SET_TOPHALF(numverts,1,0,0,PS2_GIFTAG_FLG_REGLIST,1),&(l->packet));
	AddGSData(7,PS2_GS_XYZF2,&(l->packet));
	for(int i=0; i<(numverts*2); i+=2)
	{	AddGSData(8+(i/2),PS2_GS_SETREG_XYZ((2048+x+verts[i])<<4,(2048+y+verts[i+1])<<4,0),&(l->packet)); }
	TerminateGifPacket(&(l->packet));
	
	l->x0 = x;
	l->y0 = y;
	l->x1 = x + w;
	l->y1 = y + h;
	l->packetsize = SizeGifPacket(&(l->packet));
}

void DrawLine(Line *l)
{
	ps2_dma_start_n(g_fd_gs,l->packet.begin,l->packetsize);
}


//*************************************************************************************
//										TRIANGLE
//*************************************************************************************

//The triangle coordinates v0x,v0y,etc refer to that vertex's x and y distances
//from the center of the triangle which is at x,y.  So for instance, to make a normal
//looking equilateral with the peak at the top, v0x should be 0 and v0y should
//be negative.

void InitTriangle(Triangle *t, int x, int y, int v0x, int v0y, int v1x, int v1y,
			      int v2x, int v2y, int r, int g, int b)
{
	InitGifPacket(&(t->packet),(u128 *)t->regdata);

	AddGSData(0,PS2_GIFTAG_SET_TOPHALF(1,1,0,0,PS2_GIFTAG_FLG_REGLIST,5),&(t->packet));
	AddGSData(1,(PS2_GS_PRIM | PS2_GS_RGBAQ<<4 | PS2_GS_XYZF2<<8 | 
				 PS2_GS_XYZF2<<12 | PS2_GS_XYZF2<<16),&(t->packet));
	AddGSData(2,PS2_GS_SETREG_PRIM(PS2_GS_PRIM_PRIM_TRIANGLE,0,0,0,0,0,0,0,0),&(t->packet));
	AddGSData(3,PS2_GS_SETREG_RGBAQ(r,g,b,128,0),&(t->packet));
	AddGSData(4,PS2_GS_SETREG_XYZ((2048+x+v0x)<<4,(2048+y+v0y)<<4,0),&(t->packet));
	AddGSData(5,PS2_GS_SETREG_XYZ((2048+x+v1x)<<4,(2048+y+v1y)<<4,0),&(t->packet));
	AddGSData(6,PS2_GS_SETREG_XYZ((2048+x+v2x)<<4,(2048+y+v2y)<<4,0),&(t->packet));
	TerminateGifPacket(&(t->packet));
	
	t->x = x;
	t->y = y;
	t->packetsize = SizeGifPacket(&(t->packet));
}

void InitShadedTriangle(Triangle *t, int x, int y, int v0x, int v0y, int v1x, int v1y,
			            int v2x, int v2y, int r0, int g0, int b0, int r1, int g1, 
			            int b1, int r2, int g2, int b2)
{
	InitGifPacket(&(t->packet),(u128 *)t->regdata);
	
	AddGSData(0,PS2_GIFTAG_SET_TOPHALF(1,1,0,0,PS2_GIFTAG_FLG_REGLIST,7),&(t->packet));
	AddGSData(1,(PS2_GS_PRIM | PS2_GS_RGBAQ<<4 | PS2_GS_XYZF2<<8 | PS2_GS_RGBAQ<<12 | 
				 PS2_GS_XYZF2<<16 | PS2_GS_RGBAQ<<20 | PS2_GS_XYZF2<<24),&(t->packet));
	AddGSData(2,PS2_GS_SETREG_PRIM(PS2_GS_PRIM_PRIM_TRIANGLE,1,0,0,0,0,0,0,0),&(t->packet));
	AddGSData(3,PS2_GS_SETREG_RGBAQ(r0,g0,b0,128,0),&(t->packet));
	AddGSData(4,PS2_GS_SETREG_XYZ((2048+x+v0x)<<4,(2048+y+v0y)<<4,0),&(t->packet));
	AddGSData(5,PS2_GS_SETREG_RGBAQ(r1,g1,b1,128,0),&(t->packet));
	AddGSData(6,PS2_GS_SETREG_XYZ((2048+x+v1x)<<4,(2048+y+v1y)<<4,0),&(t->packet));
	AddGSData(7,PS2_GS_SETREG_RGBAQ(r2,g2,b2,128,0),&(t->packet));
	AddGSData(8,PS2_GS_SETREG_XYZ((2048+x+v2x)<<4,(2048+y+v2y)<<4,0),&(t->packet));
	
	TerminateGifPacket(&(t->packet));
	
	t->x = x;
	t->y = y;
	t->packetsize = SizeGifPacket(&(t->packet));
}

void InitTriangleStrip(Triangle *t, int x, int y, int v0x, int v0y, int v1x, int v1y,
			           int v2x, int v2y, int *verts, int numverts, int r, int g, int b)
{
	InitGifPacket(&(t->packet),(u128 *)t->regdata);

	AddGSData(0,PS2_GIFTAG_SET_TOPHALF(1,0,0,0,PS2_GIFTAG_FLG_REGLIST,5),&(t->packet));
	AddGSData(1,(PS2_GS_PRIM | PS2_GS_RGBAQ<<4 | PS2_GS_XYZF2<<8 | 
				 PS2_GS_XYZF2<<12 | PS2_GS_XYZF2<<16),&(t->packet));
	AddGSData(2,PS2_GS_SETREG_PRIM(PS2_GS_PRIM_PRIM_TRISTRIP,0,0,0,0,0,0,0,0),&(t->packet));
	AddGSData(3,PS2_GS_SETREG_RGBAQ(r,g,b,128,0),&(t->packet));
	AddGSData(4,PS2_GS_SETREG_XYZ((2048+x+v0x)<<4,(2048+y+v0y)<<4,0),&(t->packet));
	AddGSData(5,PS2_GS_SETREG_XYZ((2048+x+v1x)<<4,(2048+y+v1y)<<4,0),&(t->packet));
	AddGSData(6,PS2_GS_SETREG_XYZ((2048+x+v2x)<<4,(2048+y+v2y)<<4,0),&(t->packet));
	AddGSData(7,0,&(t->packet));
	AddGSData(8,PS2_GIFTAG_SET_TOPHALF(numverts,1,0,0,PS2_GIFTAG_FLG_REGLIST,1),&(t->packet));
	AddGSData(9,PS2_GS_XYZF2,&(t->packet));
	for(int i=0; i<(numverts*2); i+=2)
	{	AddGSData(10+(i/2),PS2_GS_SETREG_XYZ((2048+x+verts[i])<<4,(2048+y+verts[i+1])<<4,0),&(t->packet)); }
	
	TerminateGifPacket(&(t->packet));
	
	t->x = x;
	t->y = y;
	t->packetsize = SizeGifPacket(&(t->packet));
}

void InitTriangleFan(Triangle *t, int x, int y, int v0x, int v0y, int v1x, int v1y,
			           int v2x, int v2y, int *verts, int numverts, int r, int g, int b)
{
	InitGifPacket(&(t->packet),(u128 *)t->regdata);

	AddGSData(0,PS2_GIFTAG_SET_TOPHALF(1,0,0,0,PS2_GIFTAG_FLG_REGLIST,5),&(t->packet));
	AddGSData(1,(PS2_GS_PRIM | PS2_GS_RGBAQ<<4 | PS2_GS_XYZF2<<8 | 
				 PS2_GS_XYZF2<<12 | PS2_GS_XYZF2<<16),&(t->packet));
	AddGSData(2,PS2_GS_SETREG_PRIM(PS2_GS_PRIM_PRIM_TRIFAN,0,0,0,0,0,0,0,0),&(t->packet));
	AddGSData(3,PS2_GS_SETREG_RGBAQ(r,g,b,128,0),&(t->packet));
	AddGSData(4,PS2_GS_SETREG_XYZ((2048+x+v0x)<<4,(2048+y+v0y)<<4,0),&(t->packet));
	AddGSData(5,PS2_GS_SETREG_XYZ((2048+x+v1x)<<4,(2048+y+v1y)<<4,0),&(t->packet));
	AddGSData(6,PS2_GS_SETREG_XYZ((2048+x+v2x)<<4,(2048+y+v2y)<<4,0),&(t->packet));
	AddGSData(7,0,&(t->packet));
	AddGSData(8,PS2_GIFTAG_SET_TOPHALF(numverts,1,0,0,PS2_GIFTAG_FLG_REGLIST,1),&(t->packet));
	AddGSData(9,PS2_GS_XYZF2,&(t->packet));
	for(int i=0; i<(numverts*2); i+=2)
	{	AddGSData(10+(i/2),PS2_GS_SETREG_XYZ((2048+x+verts[i])<<4,(2048+y+verts[i+1])<<4,0),&(t->packet)); }
	
	TerminateGifPacket(&(t->packet));
	
	t->x = x;
	t->y = y;
	t->packetsize = SizeGifPacket(&(t->packet));
}

void DrawTriangle(Triangle *t)
{
	ps2_dma_start_n(g_fd_gs,t->packet.begin,t->packetsize);
}

//*************************************************************************************
//										QUAD
//*************************************************************************************

void InitAlphaQuad(Quad *q, int x, int y, int w, int h, int r, int g, int b, int a)
{
	InitGifPacket(&(q->packet),(u128 *)q->regdata);
	
	AddGSData(0,PS2_GIFTAG_SET_TOPHALF(2,0,0,0,PS2_GIFTAG_FLG_PACKED,1),&(q->packet));
	AddGSData(1,PS2_GIFTAG_REGS_AD,&(q->packet));
	AddGSData(2,PS2_GS_SETREG_ALPHA(0,1,0,1,0),&(q->packet));
	AddGSData(3,PS2_GS_ALPHA_1,&(q->packet));
	// not sure why we need to set this test here, but without it
	// things don't work.  Maybe need to set at least 2 regs in packed mode?
	// I don't know.
	// this actually means--draw pixels with alpha != 0
	AddGSData(4,PS2_GS_SETREG_TEST(1,PS2_GS_TEST_ATST_NOTEQUAL,0,0,0,0,0,0),&(q->packet));
	AddGSData(5,PS2_GS_TEST_1,&(q->packet));

	AddGSData(6,PS2_GIFTAG_SET_TOPHALF(1,1,0,0,PS2_GIFTAG_FLG_REGLIST,4),&(q->packet));
	AddGSData(7,(PS2_GS_PRIM | PS2_GS_RGBAQ<<4 | PS2_GS_XYZF2<<8 | PS2_GS_XYZF2<<12),&(q->packet));
	AddGSData(8,PS2_GS_SETREG_PRIM(PS2_GS_PRIM_PRIM_SPRITE,0,0,0,1,0,0,0,0),&(q->packet));
	AddGSData(9,PS2_GS_SETREG_RGBAQ(r,g,b,a,0),&(q->packet));
	AddGSData(10,PS2_GS_SETREG_XYZ((2048+x)<<4,(2048+y)<<4,0),&(q->packet));
	AddGSData(11,PS2_GS_SETREG_XYZ((2048+x+w)<<4,(2048+y+h)<<4,0),&(q->packet));

	TerminateGifPacket(&(q->packet));
	
	q->x = x;
	q->y = y;
	q->w = w;
	q->h = h;
	q->packetsize = SizeGifPacket(&(q->packet));

}
void InitQuad(Quad *q, int x, int y, int w, int h, int r, int g, int b)
{
	InitGifPacket(&(q->packet),(u128 *)q->regdata);
	
	AddGSData(0,PS2_GIFTAG_SET_TOPHALF(1,1,0,0,PS2_GIFTAG_FLG_REGLIST,4),&(q->packet));
	AddGSData(1,(PS2_GS_PRIM | PS2_GS_RGBAQ<<4 | PS2_GS_XYZF2<<8 | PS2_GS_XYZF2<<12),&(q->packet));
	AddGSData(2,PS2_GS_SETREG_PRIM(PS2_GS_PRIM_PRIM_SPRITE,0,0,0,1,0,0,0,0),&(q->packet));
	AddGSData(3,PS2_GS_SETREG_RGBAQ(r,g,b,128,0),&(q->packet));
	AddGSData(4,PS2_GS_SETREG_XYZ((2048+x)<<4,(2048+y)<<4,0),&(q->packet));
	AddGSData(5,PS2_GS_SETREG_XYZ((2048+x+w)<<4,(2048+y+h)<<4,0),&(q->packet));
	
	TerminateGifPacket(&(q->packet));
	
	q->x = x;
	q->y = y;
	q->w = w;
	q->h = h;
	q->packetsize = SizeGifPacket(&(q->packet));
}

void MoveAlphaQuad(Quad *q,int x,int y,int w,int h)
{
	q->x = x;
	q->y = y;
	q->w = w;
	q->h = h;

 	AddGSData(10,PS2_GS_SETREG_XYZ((2048+x)<<4,(2048+y)<<4,0),&(q->packet));
	AddGSData(11,PS2_GS_SETREG_XYZ((2048+x+w)<<4,(2048+y+h)<<4,0),&(q->packet));

}

void MoveQuad(Quad *q,int x,int y,int w,int h)
{
	q->x = x;
	q->y = y;
	q->w = w;
	q->h = h;

 	AddGSData(4,PS2_GS_SETREG_XYZ((2048+x)<<4,(2048+y)<<4,0),&(q->packet));
	AddGSData(5,PS2_GS_SETREG_XYZ((2048+x+w)<<4,(2048+y+h)<<4,0),&(q->packet));

}

void DrawQuad(Quad *q)
{
	ps2_dma_start_n(g_fd_gs,q->packet.begin,q->packetsize);
}

//*************************************************************************************
//									  QUADTEX
//*************************************************************************************

//trans = 1 : all pixels of RGB = (0,0,0) will be transparent
//trans = 0 : no transparency

void InitQuadTex(QuadTex *q, int x, int y, int w, int h, int trans, int filter, Texture *t)
{
	InitGifPacket(&(q->packet),(u128 *)q->regdata);
	
	AddGSData(0,PS2_GIFTAG_SET_TOPHALF(5,0,0,0,PS2_GIFTAG_FLG_PACKED,1),&(q->packet));
	AddGSData(1,PS2_GIFTAG_REGS_AD,&(q->packet));
	AddGSData(2,PS2_GS_SETREG_TEXA(128,trans,0),&(q->packet));
	AddGSData(3,PS2_GS_TEXA,&(q->packet));
	AddGSData(4,PS2_GS_SETREG_TEST(trans,PS2_GS_TEST_ATST_NOTEQUAL,0,0,0,0,0,0),&(q->packet));
	AddGSData(5,PS2_GS_TEST_1,&(q->packet));
	AddGSData(6,PS2_GS_SETREG_TEX0(t->image_adr,(t->w+63)/64,t->psm,
	                               t->log2width,t->log2height,1,1,
	                               t->clut_adr,t->clut_psm,0,0,t->clut_load),&(q->packet));
	AddGSData(7,PS2_GS_TEX0_1,&(q->packet));
	AddGSData(8,PS2_GS_SETREG_TEX1_1(0,0,filter,filter,0,0,0),&(q->packet));
	AddGSData(9,PS2_GS_TEX1_1,&(q->packet));
	AddGSData(10,PS2_GS_SETREG_CLAMP(1,1,0,t->w,0,t->h),&(q->packet));
	AddGSData(11,PS2_GS_CLAMP_1,&(q->packet));
	AddGSData(12,PS2_GIFTAG_SET_TOPHALF(1,1,0,0,PS2_GIFTAG_FLG_REGLIST,6),&(q->packet));
	AddGSData(13,(PS2_GS_PRIM | PS2_GS_RGBAQ<<4 | PS2_GS_UV<<8 | PS2_GS_XYZF2<<12 |
				  PS2_GS_UV<<16 | PS2_GS_XYZF2<<20),&(q->packet));
	AddGSData(14,PS2_GS_SETREG_PRIM(PS2_GS_PRIM_PRIM_SPRITE,0,1,0,0,0,1,0,0),&(q->packet));
	AddGSData(15,PS2_GS_SETREG_RGBAQ(255,0,0,128,0),&(q->packet));
	AddGSData(16,PS2_GS_SETREG_UV(0,0),&(q->packet));
	AddGSData(17,PS2_GS_SETREG_XYZ((2048+x)<<4,(2048+y)<<4,0),&(q->packet));
	AddGSData(18,PS2_GS_SETREG_UV(t->w<<4,t->h<<4),&(q->packet));
	AddGSData(19,PS2_GS_SETREG_XYZ((2048+x+w)<<4,(2048+y+h)<<4,0),&(q->packet));

	TerminateGifPacket(&(q->packet));
	
	q->tex = t;
	q->x = x;
	q->y = y;
	q->w = w;
	q->h = h;
	q->packetsize = SizeGifPacket(&(q->packet));
}

void SetQuadTexFilter(QuadTex *q,int filter)
{
	AddGSData(8,PS2_GS_SETREG_TEX1_1(0,0,filter,filter,0,0,0),&(q->packet));
}

void MoveQuadTex(QuadTex *q,int x,int y,int w,int h)
{
	q->x = x;
	q->y = y;
	q->w = w;
	q->h = h;

 	AddGSData(17,PS2_GS_SETREG_XYZ((2048+x)<<4,(2048+y)<<4,0),&(q->packet));
	AddGSData(19,PS2_GS_SETREG_XYZ((2048+x+w)<<4,(2048+y+h)<<4,0),&(q->packet));
}

void SetQuadTexUV(QuadTex *q,int uvx0,int uvy0,int uvx1,int uvy1)
{
	AddGSData(16,PS2_GS_SETREG_UV(uvx0<<4,uvy0<<4),&(q->packet));
	AddGSData(18,PS2_GS_SETREG_UV((uvx1)<<4,(uvy1)<<4),&(q->packet));
}
void DrawQuadTex(QuadTex *q)
{
	ps2_dma_start_n(g_fd_gs,q->packet.begin,q->packetsize);
}


//*************************************************************************************
//									     FONT
//*************************************************************************************

//spacing is the distance between letters, negative numbers
//usually work best

void InitFont(Font *f, Texture *t, int spacing)
{
	InitGifPacket(&(f->packet),(u128 *)f->regdata);

	f->tex = t;
	f->x = 2048;
	f->y = 2048;
	f->w = t->w/16;
	f->h = t->h/8; // 1 to 1 pixel-wise with the original
	f->space = spacing;	
}

//trans = 1 : all pixels of RGB = (0,0,0) will be transparent
//trans = 0 : no transparency

void DrawString(const char *str, int x, int y, Font *f,bool trans)
{
	static char c,lc;
	static const char *l;
	static unsigned short *tc;
	static int uvx0,uvy0,uvx1,uvy1,length;
	
	x += 2048;
	y += 2048;
	length = 0;
	c = *str;
	l = str;
	lc = *l;
	
	while(lc)
	{	length++;
		l++;
		lc = *l;
	}
	
	ResetGifPacket(&(f->packet));
	
	AddGSData(0,PS2_GIFTAG_SET_TOPHALF(5,0,0,0,PS2_GIFTAG_FLG_PACKED,1),&(f->packet));
	AddGSData(1,PS2_GIFTAG_REGS_AD,&(f->packet));
	AddGSData(2,PS2_GS_SETREG_TEXA(128,trans,0),&(f->packet));
	AddGSData(3,PS2_GS_TEXA,&(f->packet));
	AddGSData(4,PS2_GS_SETREG_TEST(1,PS2_GS_TEST_ATST_NOTEQUAL,0,0,0,0,0,0),&(f->packet));
	AddGSData(5,PS2_GS_TEST_1,&(f->packet));
	AddGSData(6,PS2_GS_SETREG_TEX0(f->tex->image_adr,(f->tex->w+63)/64,f->tex->psm,
                                   f->tex->log2width,f->tex->log2height,1,1,
                                   f->tex->clut_adr,f->tex->clut_psm,0,0,f->tex->clut_load),&(f->packet));                                          
	AddGSData(7,PS2_GS_TEX0_1,&(f->packet));
	AddGSData(8,PS2_GS_SETREG_TEX1_1(0,0,0,0,0,0,0),&(f->packet));
	AddGSData(9,PS2_GS_TEX1_1,&(f->packet));
	AddGSData(10,PS2_GS_SETREG_CLAMP(1,1,0,f->tex->w,0,f->tex->h),&(f->packet));
	AddGSData(11,PS2_GS_CLAMP_1,&(f->packet));
	
	AddGSData(12,PS2_GIFTAG_SET_TOPHALF(length,1,0,0,PS2_GIFTAG_FLG_REGLIST,6),&(f->packet));
	AddGSData(13,(PS2_GS_PRIM | PS2_GS_RGBAQ<<4 | PS2_GS_UV<<8 | PS2_GS_XYZF2<<12 | 
				  PS2_GS_UV<<16 | PS2_GS_XYZF2<<20),&(f->packet));
	
	for(int i=0; i<length; i++)
	{	
		tc = &fixed_tc[c*4];
		uvx0 = *tc++;
		uvy0 = *tc++;
		uvx1 = *tc++;
		uvy1 = *tc++;
	
		AddGSData(14+i*6,PS2_GS_SETREG_PRIM(PS2_GS_PRIM_PRIM_SPRITE,0,1,0,0,0,1,0,0),&(f->packet));
		AddGSData(15+i*6,PS2_GS_SETREG_RGBAQ(255,0,0,128,0),&(f->packet));
		AddGSData(16+i*6,PS2_GS_SETREG_UV(uvx0<<4,(uvy0+2)<<4),&(f->packet));
		AddGSData(17+i*6,PS2_GS_SETREG_XYZ(x<<4,y<<4,0),&(f->packet));
		AddGSData(18+i*6,PS2_GS_SETREG_UV((uvx1+1)<<4,(uvy1+1)<<4),&(f->packet));
		AddGSData(19+i*6,PS2_GS_SETREG_XYZ((x+f->w)<<4,(y+f->h)<<4,0),&(f->packet));
		
		// Get next character
		str++;
		c = *str;
		x += (f->w + f->space);
	}
		
	TerminateGifPacket(&(f->packet));
	ps2_dma_start_n(g_fd_gs,f->packet.begin,SizeGifPacket(&(f->packet)));
}
