#include "texture.h"

extern int g_textop64; 
extern int g_psm;	  //both from main.c

Texture::Texture()
{
	image_data = NULL;
	clut_data = NULL;
	w,h = 0;
	log2width,log2height = 0;
	image_adr = 0;
	psm = 0;
	clut_adr = 0;
	clut_psm = 0;
	clut_load = 0;

}

Texture::~Texture()
{
	if(image_data != NULL)
		free(image_data);
	if(clut_data != NULL)
		free(clut_data);
}

// image is not uploaded to GS, you need to do that yourself
// with ResendImageToGS(), and palette with ResendPaletteToGS() 
bool Texture::Create8BitPalettized(int width,int height)
{
	clut_data = (Byte *)memalign(16,256*4);

	w = width;
	h = height;

	image_data = (Byte *)memalign(16,w*h);

	log2width = Log2(w);
	log2height = Log2(h);

	if (w != (1 << log2width) || h != (1 << log2height))
		printf("invalid created texture size %i x %i\n");

	psm = PS2_GS_PSMT8;
	clut_load = 1;
	clut_psm = PS2_GS_PSMCT32;

	//Set image data address to current GSRAM pointer
	clut_adr = g_textop64;
	image_adr = clut_adr + 256/64;

	
	int gswordsNeeded;
	
	// image data + clut in 32 bit units
	gswordsNeeded = (w * h)/4 + 256;
	
	int dtex64 = (gswordsNeeded + 63) / 64;
	
	int cbUsed = g_textop64*64*4;
	//printf("GS RAM: %d used, %d free, %d needed for texture %s\n",
	//		cbUsed, 4*1024*1024 - cbUsed, gswordsNeeded*4,filename);
	
	if (g_textop64 + dtex64 >= 16*1024)
	{
		printf("GS RAM full %d used, %d free, %d needed for texture.\n",
				 cbUsed, 4*1024*1024 - cbUsed, gswordsNeeded*4);
		
		return false;	// no room
	}
	
	//Advance GSRAM pointer so as not to overwrite it the next time
	g_textop64 += dtex64;
	
	return true;
}

void Texture::SetPaletteEntry(Byte index,Byte r,Byte g,Byte b,Byte a)
{
	if (clut_data == NULL)
	{
		printf("can't set palette on a non-palettized texture");
		return;
	}

	index = swapBits3and4(index);

	clut_data[4*index] = r;
	clut_data[4*index+1] = g;
	clut_data[4*index+2] = b;
	clut_data[4*index+3] = a;
}

int Texture::LoadBitmap(char *filename)
{
	FILE *file;
	BitmapFile bitmap;
	int imagesize = 0; //size in bytes of actual image data
	
	file = fopen(filename,"rb");
	if(file == NULL)
	{	printf("%s: Could not open file\n",filename);
		return 0;
	}
	
	//I've had a real problem with loading in these first few
	//items, so read them in separately.
	//fread(&(bitmap.bitmapfileheader),14,1,file); //should be this?
	fread(&(bitmap.bitmapfileheader.bfType),2,1,file);
	fread(&(bitmap.bitmapfileheader.bfSize),4,1,file);
	fread(&(bitmap.bitmapfileheader.bfReserved1),2,1,file);
	fread(&(bitmap.bitmapfileheader.bfReserved2),2,1,file);
	fread(&(bitmap.bitmapfileheader.bfOffBits),4,1,file);
	
	if(bitmap.bitmapfileheader.bfType!=BITMAP_ID)
	{
		printf("%s: Not a valid bitmap file\n",filename);
		printf("%s: bfType is %d, should be %d\n",filename,bitmap.bitmapfileheader.bfType,BITMAP_ID);
		fclose(file);
		return 0;
	}
	
	//Read in entire bitmapinfoheader
	fread(&(bitmap.bitmapinfoheader),sizeof(Byte),(int)(sizeof(BitmapInfoHeader)),file);
		
	if(bitmap.bitmapinfoheader.biBitCount == 8)
	{	
		//Allocate CLUT data
		clut_data = (Byte *)memalign(16,256*4);
					
		//Read in CLUT data
		fread(clut_data, 1, 256*4,file);
		
		Byte temp;
		for(int i=0; i<256; i++)
		{
			//Swap B and R values to make RGB instead of BGR
			temp = clut_data[i*4];
			clut_data[i*4] = clut_data[i*4+2];
			clut_data[i*4+2] = temp;
			
			//Set alpha value to 0x00 if RGB=(0,0,0) for transparency,
			//else 0x80
			if((clut_data[i*4]==0) && (clut_data[i*4+1]==0) && clut_data[i*4+2]==0)
				clut_data[i*4+3] = 0x00;
			else clut_data[i*4+3] = 0x80;
		}
	}
	
	//Find beginning of actual image data
	fseek(file,-(int)(bitmap.bitmapfileheader.bfOffBits),SEEK_SET);
	
	if((bitmap.bitmapinfoheader.biBitCount==8) || (bitmap.bitmapinfoheader.biBitCount == 16)
	    || (bitmap.bitmapinfoheader.biBitCount == 24))
	{
		//I've had trouble getting the biSizeImage to be
		//read in correctly, so calculate how much image data
		//to read in another way
		if(bitmap.bitmapinfoheader.biSizeImage!=0)
			imagesize = bitmap.bitmapinfoheader.biSizeImage;
		else
			imagesize = bitmap.bitmapfileheader.bfSize - bitmap.bitmapfileheader.bfOffBits;
		
		//Allocate memory for image data
		image_data = (Byte *)memalign(16,imagesize);
		
		if(!image_data)
		{
			printf("%s: malloc failed\n",filename);
			fclose(file);
			return 0;
		}
		
		//Read in image data
		fread(image_data,sizeof(Byte),imagesize,file);
		
		if(image_data == NULL)
		{	
			printf("%s: image_data read failed\n",filename);
			fclose(file);
			return 0;
		}
	}
	else
	{
		printf("%s: Not 8, 16 or 24-bit bitmap\n",filename);
		printf("%s: Bit count = %d\n",filename,bitmap.bitmapinfoheader.biBitCount);
		fclose(file);
		return 0;
	}
	
	fclose(file);
	
	//Set Texture member variables
	w = bitmap.bitmapinfoheader.biWidth;
	h = bitmap.bitmapinfoheader.biHeight;
	log2width = Log2(w);
	log2height = Log2(h);
	if (w != (1 << log2width) || h != (1 << log2height))
		printf("invalid texture size %i x %i, %s\n",w,h,filename);

	switch(bitmap.bitmapinfoheader.biBitCount)
	{
		case 8:  
		{
			//printf("loading 8 bit bitmap %s, %i x %i\n",filename,w,h);
			psm = PS2_GS_PSMT8;
			clut_load = 1;
			clut_psm = PS2_GS_PSMCT32;
			break;
		}
		case 16: 
		{
			//printf("loading 16 bit bitmap %s\n",filename);
			psm = PS2_GS_PSMCT16;
			clut_load = 0;	//no CLUT
			clut_psm = 0;
			break;
		}
		case 24:
		{
			//printf("loading 24 bit bitmap %s\n",filename);
			psm = PS2_GS_PSMCT24;
			clut_load = 0;	//no CLUT
			clut_psm = 0;
		}
	}	
	
	//Most bitmaps are up-side down, so flip right-side up
	if(bitmap.bitmapinfoheader.biHeight>0)
	{	
		if(FlipBitmap(image_data,w * (bitmap.bitmapinfoheader.biBitCount/8),h) != 1)
		{
			printf("%s: FlipBitmap failed\n",filename);
			return 0;
		}
	}
	
	//Bitmaps are encoded in BGR format, so swap B and R to make RGB
	if(bitmap.bitmapinfoheader.biBitCount!=8)
	{	for (int index_y = 0; index_y < h; index_y++)
		{
			for (int index_x = 0; index_x < w; index_x++)
    		{
        		Byte  blue  = image_data[index_y*w*3 + index_x*3 + 0];
				image_data[index_y*w*3 + index_x*3 + 0] = image_data[index_y*w*3 + index_x*3 + 2];
      			image_data[index_y*w*3 + index_x*3 + 2] = blue;
			}
		}
	}
	else	//8-bit bitmaps need to be adjusted for CLUT entries
	{
		for(int i=0; i<w*h; i++)
			image_data[i] = swapBits3and4(image_data[i]);
	}
	
	
	//**************************************************************
	//**************************AllocateGS**************************
	//**************************************************************
	//Checks to see if there's enough GSRAM
	//Advances current GSRAM pointer(g_textop64) accordingly
	
	//Set image data address to current GSRAM pointer
	image_adr = g_textop64;
	
	int gswordsNeeded;
	
	if(psm == PS2_GS_PSMT8)
		gswordsNeeded = (w * h)/2 + 256; //make room for CLUT
	else                                 //should be (w*h)/4?
		gswordsNeeded = w * h;
	
	int dtex64 = (gswordsNeeded + 63) / 64;
	
	int cbUsed = g_textop64*64*4;
	//printf("GS RAM: %d used, %d free, %d needed for texture %s\n",
	//		cbUsed, 4*1024*1024 - cbUsed, gswordsNeeded*4,filename);
	
	if (g_textop64 + dtex64 >= 16*1024)
	{
		printf("GS RAM full %d used, %d free, %d needed for texture %s\n",
			cbUsed, 4*1024*1024 - cbUsed, gswordsNeeded*4,filename);
		
		return 0;	// no room
	}
	
	//Advance GSRAM pointer so as not to overwrite it the next time
	g_textop64 += dtex64;
	
	//**************************************************************
	//************************CopyToTextureRAM**********************
	//**************************************************************	
	//Copy CLUT(if necessary) and image data to GSRAM using
	//function ps2_gs_load_image
	
	ps2_image img;

	if(psm == PS2_GS_PSMT8)
	{
		clut_adr = image_adr;
		image_adr += 256/64;

		ps2_gs_set_image(&img,clut_adr,1,clut_psm,
					 	 0, 0, 16, 16, clut_data);
	
		if (ps2_gs_load_image(&img) == -1)
		{	printf("%s: ps2_gs_load_image of clut failed\n",filename);
			return 0;
		}
		
		ps2_gs_set_image(&img,image_adr, (w+63)/64, psm,
					     0, 0, w, h, image_data);
	
		if (ps2_gs_load_image(&img) == -1)
		{	printf("%s: ps2_gs_load_image failed\n",filename);
			return 0;
		}
	}
	else
	{
		ps2_gs_set_image(&img,image_adr, (w+63)/64, psm,
					 0, 0, w, h, image_data);
	
		if (ps2_gs_load_image(&img) == -1)
		{	printf("%s ps2_gs_load_image failed\n",filename);
			return 0;
		}
	}

	return 1;
}

void Texture::ResendPaletteToGS()
{
	ps2_image img;

	ps2_gs_set_image(&img,clut_adr,1,clut_psm,
						  0, 0, 16, 16, clut_data);
	
	if (ps2_gs_load_image(&img) == -1)
		printf("ps2_gs_load_image of clut failed\n");
		
}
void Texture::ResendImageToGS()
{
	ps2_image img;

	ps2_gs_set_image(&img,image_adr, (w+63)/64, psm,
					     0, 0, w, h, image_data);
	
	if (ps2_gs_load_image(&img) == -1)
		printf("ps2_gs_load_image failed\n");

}

int FlipBitmap(Byte *image, int bytes_per_line, int height)
{
	Byte *buffer;
	int index;

	if (!(buffer = (Byte *)malloc(bytes_per_line*height)))
   	{	printf("In FlipBitmap: malloc buffer failed\n");
   		return(0);
   	}
	
	memcpy(buffer,image,bytes_per_line*height);
	
	for (index=0; index < height; index++)
    	memcpy(&image[((height-1) - index)*bytes_per_line],&buffer[index*bytes_per_line], bytes_per_line);

	free(buffer);

	return 1;
}

Word Log2(Word n)
{
	Word log=0;
	
	n--;
	while(n>0)
	{	n>>=1;
		log++;
	}
	
	return log;
}

//From CFYC learning03: http://cfyc.hikey.org/
unsigned int swapBits3and4(unsigned int value) // works on 32bit value
{
	return (value&0xe7e7e7e7) | ((value&0x08080808)<<1) | ((value&0x10101010)>>1);
}
